#    Lunar Rover Expedition - A fun adventure game.
#    Copyright (C) 2013 Jannik Haberbosch
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Importion of modules
import os.path
import string

import pygame
pygame.init()

from code.other.constants import LEVELS_DIR_PATH
from code.classes.Sample import Sample
from code.classes.Level_Clock import Level_Clock
from code.classes.Level_Meter import Level_Meter
from code.classes.Static_Enemy import Static_Enemy
from code.classes.Kamikaze_Enemy import Kamikaze_Enemy
from code.classes.Moving_Enemy import Moving_Enemy


class Level( object ):
	"""
	This is the class for a level of the game.

	A level consists of objects that are drawn onto the screen.
	The entire level is _centered_ on the window.
	"""


	def __init__( self, level_map_file, map_center_position, lvl_nr, hash_table, tile_size, window_dimensions, energy_production, level_clock_data, level_meter_data ):
		"""
		Parameters
			level_map_file		string
				The file that contains the map for the current level.
			map_center_position	( int, int ) or [ int, int ]
				Defines what coords constitute as the center position of the level.
			lvl_nr				int
				The level number.
			hash_table			dict { string, string }
				Each key is a character that represents an image, that in turn represents an object.
			tile_size			int
				Defines the width and height of a single tile of the map.
			window_dimensions	( int, int ) or [ int, int ]
				How big is the window?
			energy_production	int
				Defines the amount of energy that can be produced by the solar panel in a frame, so it can recharge its batteries.
			level_clock_data	dict
				Contains data for the class Level_Clock.
			level_meter_data	dict
				Contains data for the class Level_Meter.
		"""

		# I store the parameters, as I might need them later on again.
		self._level_map_file = level_map_file
		self._map_center_position = map_center_position
		self._lvl_nr = lvl_nr # Which level are we on?
		self._hash_table = hash_table
		self._tile_size = tile_size
		self._window_dimensions = window_dimensions
		self._energy_production = energy_production

		# Here I set a few variables to 0.  These values will later be overwritten.
		# These variables define the width and height of a level in pixels.
		self._level_width = 0
		self._level_height = 0


		# A level has 3 differen sprite groups, which will all organized in a superior sprite group.
		# These 3 sprite groups are:
			# self._samples - Contains the samples which the user has to collect with his lunar rover in a level.
			# self._obstacles - Contains the obstacles which makes the life of the user harder during a level.
			# self._scenery - Contains scnenery that does not affect the user, but just add to the environment, by enriching it with detail.
			# self._shooting_enemies - Contains all the enemies that can shoot.
		self._samples = pygame.sprite.Group()
		self._obstacles = pygame.sprite.Group() # Includes moving aliens / enemies.
		self._scenery = pygame.sprite.Group()
		self._shooting_enemies = pygame.sprite.Group()

		# Create all objects for this level.		
		self.set_up_level_objects()

		# These 3 sprite groups are then, as mentioned above, stored in a single sprite group.
		# This makes it easier to update and manage the sprites altogether.
		self._all_level_objects = pygame.sprite.Group()
		self._all_level_objects.add( self._samples, self._obstacles, self._scenery, self._shooting_enemies )


		# In each level, a certain amount of samples have to be collected.
		self._n_samples_to_collect = len( self._samples )


		# Create the level clock.
		self._level_clock = Level_Clock( **level_clock_data )

		# Create the level meter.
		self._level_meter = Level_Meter( **level_meter_data )

	def set_up_level_objects( self ):
		"""
		A function that sets up the objects of a level.

		There are 3 groups of objects; that means 3 classes:

			self._scenery_objects:
				No function themselves; just serving decorating purposes.
			self._enemies:
				Are an obstacles for the player.  They are able to move.
			self._samples:
				The samples that have to be detected and collected by the lunar rover.
			self._shooting_enemies
				These are the enemies that can shoot.

		Each level object has a width and height of 40px.
		"""

		# I begin by reading the content of the entire level and storing its lines into a list.
		level_rows = self.read_level()

		# I figure out the entire width and height of that level.
		self.get_level_dimensions( level_rows )

		# Define from which top-left coord the tiles should be placed from.
		left_start, top_start = self.set_placement_top_left_start()
		# I keep a copy of the original left starting coordinate position, because every new row starts from the same left starting position.
		original_left_start = left_start
		
		# I generate the map.  I cycle through each row and through each item of a row (column, that is).
		for row in level_rows:
			for col in row:
			
				# I assign the current left-top position to the selected item of self._hash_table.
				self._hash_table[col]["left_top_position"] = ( left_start, top_start )
				
				# Now the tiles are generated.
				if col == "*":
					self._samples.add( Sample( **self._hash_table[col] ) )
				elif col == "+":
					# Add a static enemy.
					static_enemy = Static_Enemy( **self._hash_table[col] )
					self._shooting_enemies.add( static_enemy )
					self._obstacles.add( static_enemy )
				elif col == "x":
					# Add a kamikaze enemy.
					kamikaze_enemy = Kamikaze_Enemy( **self._hash_table[col] )
					self._obstacles.add( kamikaze_enemy )
				elif col == "y":
					# Add a moving enemy that can shoot aswell.
					rover_enemy = Moving_Enemy( **self._hash_table[col] )
					self._shooting_enemies.add( rover_enemy )
					self._obstacles.add( rover_enemy )
					
				# The following tile is located next to it, on the right, touching it.
				left_start += self._tile_size

			# A new row starts from the left again.
			left_start = original_left_start

			# The new row is located just below the old row.
			top_start += self._tile_size


	def read_level( self ):
		"""
		Open the level file, read its contents and split the lines into rows.
		These rows are then returned.

		Return value
			rows	int
				The number of rows found in a level file.
		"""

		# I open the level file in reading mode.
		file_handle = open( os.path.join( LEVELS_DIR_PATH, self._level_map_file ), "r" )
		# I read the entire contents and store them in 'level_contents'.
		level_contents = file_handle.read()
		# I close the file.
		file_handle.close()

		# Here the content just read is split into lines.  Each line is an element in a list.
		# The elements are stored in "level_rows".
		# The function string.split() returns an empty line at the end.
		level_rows = string.split( level_contents, "\n" )

		return level_rows


	def get_level_dimensions( self, level_rows ):
		"""
		How large is the level in pixels ?
		This is what I figure out with this function.

		Parameters
			level_rows	( str, str, ... ) or [ str, str ]
				The number of rows of a level.
		"""

		# Find the length of the longest string in "level_rows".  Or max-cols.
		cols = len( max( level_rows, key=len ) )

		# Store the amount of rows.  Or max-rows.
		# I subtract "1", because the function string.split(), that has been used previously, always returns an additional, empty row.
		rows = len( level_rows ) - 1

		# The width of the level in pixels.
		self._level_width = cols * self._tile_size
		# The height of the level in pixels.
		self._level_height = rows * self._tile_size


	def set_placement_top_left_start( self ):
		"""
		From which top-left position should the tiles be placed later on?  What coords should be the starting point?
		This is what I figure out here.

		Return values
			left_start	int
				The left coordinate.
			top_start	int
				The top coordinate.
		"""

		left_start = ( ( self._level_width - 640 ) / 2 ) * -1
		top_start = ( ( self._level_height - 480 ) / 2 ) * -1

		return left_start, top_start


	def all_samples_collected( self ):
		"""
		Checks if all samples of this level have been collected by the lunar rover.  If that is the case, then 'True' is returned; otherwise, 'False' is returned.

		Return value
			True
				All samples of this level have been collected.
			False
				There are still some samples to collect in this level.
		"""

		if len( self._samples ) == 0:
			return True
		return False


	def update_level_clock( self ):

		self._level_clock.update_time()
