#    Lunar Rover Expedition - A fun adventure game.
#    Copyright (C) 2013 Jannik Haberbosch
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Importion of modules
import math

import pygame
pygame.init()

from code.classes.Animated_Sprite import Animated_Sprite
from code.classes.Shot import Shot
from code.classes.Lunar_Rover_Health import Lunar_Rover_Health
from code.classes.Lunar_Rover_Sample_Meter import Lunar_Rover_Sample_Meter
from code.classes.Lunar_Rover_Energy_Meter import Lunar_Rover_Energy_Meter

from code.other.helpers import load_sound

class Lunar_Rover( Animated_Sprite ):
	"""
	This is the class for the lunar rover, which the player controls during the game.
	"""


	def __init__( self, animation_image_file_names, center_xy_start_pos, animation_intervals, keys, speed, action_states, angle, degree_amount, shot, collision_rect_data, vehicle_health_data, sample_meter_data, energy_meter_data, energy_consumption, energy_production, shot_damage, sound_shot_filename, sound_sample_collected_filename ):
		"""
		Parameters
			animation_image_file_names	( string, string, ... )
				A list of image file names of images that will be used for the animation.
			center_xy_start_pos			( int, int )
				Defines the starting positon of the lunar rover.  The coords represent the center of it.
			animation_intervals			( int, int, ... )
				A list that contains the animation intervals.  They control at what speed the animation runs.
			keys						dict { pygame key: string, ... }
				A dictionary that consists of key:value pairs.  The key is a pygame Button and the string tells what to do if that key is pressed.
			speed						int
				Defines the speed of the lunar rover.
			action_states				dict { string, bool }
				A dictionary which tells whether a specific again should be performed; such as moving forward or turning to the right.
			angle						int
				Defines the starting angle of the lunar rover.  Defines where it has to head to when the player drives it.
			degree_amount				int
				Defines by how much the angle should be altered (used as both negative and positive value) when the left arrow key or the right arrow key has been pressed.
			shot						dictionary
				This dictionary contains all the necessary data to create a shot that can be fired by the vehicle.  It is used in conjunction with the class "Shot".  The reason it is passed to this class is that it belongs to the lunar rover and the angle of the lunar rover has to be known in order to fire the projectile in the right direction.
			collision_rect_data			( int, int ) or [ int, int ]
				Coords for a rectangle that is used for collision between the lunar rover and other level objects.
			vehicle_health_data			dict
				Contains data for the vehicle health meter.
			sample_meter_data			dict
				Contains data for the samples meter.
			energy_meter_data			dict
				Contains data for the energy meter.
			energy_consumption			dict
				Defines what actions need how much energy.
			energy_production			int
				How much energy can be produced by the solar panel in a frame?
			shot_damage					int
				Defines how much damage a shot fired by the lunar rover does to an enemy.
			sound_shot_filename			int
				The filename of the sound that will be played if the player fires a shot.
			sound_sample_collected_filename	string
				""
		"""
		
		super( Lunar_Rover, self ).__init__( animation_image_file_names, animation_intervals )

		# Store the parameters, as I might need to reuse them later on.
		self._center_xy_start_pos = center_xy_start_pos
		self._keys = keys
		self._speed = speed
		self._action_states = action_states
		self._angle = angle
		self._degree_amount = degree_amount
		self._shot_data = shot
		self._collision_rect_data = collision_rect_data

		self._energy_consumption_drive = energy_consumption["drive"]
		self._energy_consumption_shoot = energy_consumption["shoot"]
		self._energy_consumption_radar = energy_consumption["radar"]
		self._energy_consumption_turn = energy_consumption["turn"]

		self._energy_production	= energy_production
		self._energy_meter_text_update_interval_counter = 0

		self._shot_damage = shot_damage

		# Create the vehicle health data object
		self._vehicle_health = Lunar_Rover_Health( **vehicle_health_data )

		# Create the vehicle sample meter object
		self._sample_meter = Lunar_Rover_Sample_Meter( **sample_meter_data )

		# Create the vehicle's energy meter object
		self._energy_meter = Lunar_Rover_Energy_Meter( **energy_meter_data )

		# Not passed as parameters
		self._velocity = [ 0, 0 ]
		self._current_speed = 0
		self._shots = pygame.sprite.Group()
		self._shot_already_fired = False # Prevent continuous shots being fired if the spacebar is held down.
		
		# Set up the sprite.
		self.rect = self.image.get_rect( center = self._center_xy_start_pos )

		# Set up the rect used for collisions.
		self._collision_rect = pygame.Rect( self._collision_rect_data )
		self._collision_rect.center = self._center_xy_start_pos

		# Sounds
		self._sound_shot = load_sound( sound_shot_filename )
		self._sound_sample_collected = load_sound( sound_sample_collected_filename )


	def update( self, pressed_keys ):
		"""
		Updates the lunar rover.

		Parameters
			pressed_keys		pygame event dict
				A dict of keys that shows whether they have been pressed or not.

		Return value
			Boolean
				Returns 'True' if the lunar rover has been destroyed.  Otherwise, 'False' is returned.
		"""

		self.update_animation()
		self.check_keys( pressed_keys )
		self.check_action_states()
		self.calculate_xy_velocity()		

		# Note: This is a "hack" in order to skip resetting a previously rotated image.
		# Not very nice, but works like a charm.  I do this because altering "Animated_Sprite"
		# would cause too much trouble now.
		self.rotate()

		# Update any active shots.
		self._shots.update()

		# Keep shots straight.
		for s in self._shots:
			s.enemy_keep_shot_straight( self._velocity )


		# Update the health of the vehicle if necessary.  Always call this function.
		self._vehicle_health.update_text()

		# Check if the lunar rover has been defeated.
		if self._vehicle_health.is_defeated():
			return True

		# The solar panel helps to collect more energy.
		self._energy_meter.energy_increase_by_solar_panel()

		# Check if the radar is turned on, and whether there is enough energy to leave it turned on.
		if self._terrain_radar_ref._radar_is_on:
			if self.enough_energy( self._energy_consumption_radar ):
				self._energy_meter.decrease_energy( self._energy_consumption_radar )
			else:		
				self._terrain_radar_ref._radar_is_on = False

		return False

		
	def check_keys( self, pressed_keys ):
		"""
		This method checks the status of the keys and checks, if the lunar rover should do something.

		Parameters
			pressed_keys	pygame event list.
				A list of events that have happened in the current frame.
		"""

		for key_value_pair in self._keys.iteritems():
			if pressed_keys[key_value_pair[0]]:
				self._action_states[key_value_pair[1]] = True
			else:
				# Make sure the user can shoot again after he has lifted the space bar.
				if key_value_pair[1] == "shoot":
					self._shot_already_fired = False
				self._action_states[key_value_pair[1]] = False


	def check_action_states( self ):
		"""
		Checks if specific actions should be executed, such as moving the lunar rover.
		"""

		# Holding down the spacebar will only fire one projectile.  It has to be lifted first, in order to fire a projectile again.
		if self._action_states["shoot"] and not self._shot_already_fired:
			# Check if there is enough energy available to fire a projectile.
			if self.enough_energy( self._energy_consumption_shoot ):
				self.fire_shot()
				self._sound_shot.play()
				self._energy_meter.decrease_energy( self._energy_consumption_shoot )
				self._shot_already_fired = True

		if self._action_states["turn_left"]:
			# Check if there is enough energy available to turn.
			if self.enough_energy( self._energy_consumption_turn ):
				self._energy_meter.decrease_energy( self._energy_consumption_turn )			
				self.change_angle()
			
		if self._action_states["turn_right"]:
			# Check if there is enough energy available to turn.
			if self.enough_energy( self._energy_consumption_turn ):	
				self._energy_meter.decrease_energy( self._energy_consumption_turn )				
				self.change_angle( True )
			
		if self._action_states["drive_forward"]:
			# Check if there is enough energy available to move forward.
			if self.enough_energy( self._energy_consumption_drive ):
				self._energy_meter.decrease_energy( self._energy_consumption_drive )
				self._current_speed = self._speed
			else:
				self._current_speed = 0				
		elif self._action_states["drive_backward"]:
			# Check if there is enough energy available to move backward.
			if self.enough_energy( self._energy_consumption_drive ):		
				self._energy_meter.decrease_energy( self._energy_consumption_drive )		
				self._current_speed = -self._speed
			else:
				self._current_speed = 0
		else:
			self._current_speed = 0


	def rotate( self ):
		"""
		Rotates the lunar rover by a certain amount of degrees.
		It can either be rotated to the left, or to the right.
		"""

		# Store a copy of the current image.  Never rotate a previously rotated image.
		# Always use the original.
		copied_image = self._animation_images[self._shown_image].copy()
		# Store the current center coordinates of the lunar rover.
		copied_center_xy = self.rect.center

		# Rotate the image now.
		rotated_image = pygame.transform.rotate( copied_image, self._angle )
		self.image = rotated_image
		self.rect = self.image.get_rect( center = copied_center_xy )


	def change_angle( self, negative=False ):
		"""
		Changes the angle of the lunar rover.
		If negative is True, then the degree value will be inversed.
		"""

		if negative:
			if self._angle <= 360 and self._angle > ( 0 + self._degree_amount ):
				self._angle -= self._degree_amount
			else:
				self._angle = 360

		else:
			if self._angle >= 0 and self._angle < ( 360 - self._degree_amount ):
				self._angle += self._degree_amount
			else:
				self._angle = 0


	def calculate_xy_velocity( self ):
		"""
		Calculates the velocity along the x and y axes.
		These values are then required by all other game objects, in order to get the lunar rover
		"moving".
		"""

		velocity_x = self._current_speed * math.sin( math.radians( self._angle ) )
		velocity_y = self._current_speed * math.cos( math.radians( self._angle ) )

		self._velocity = [ velocity_x, velocity_y ]


	def fire_shot( self ):
		"""
		Fires a shot.
		"""

		# Update the angle, otherwise the shot will be fired in the wrong direction.
		self._shot_data["angle"] = self._angle
		self._shots.add( Shot( **self._shot_data ) )


	def enough_energy( self, energy_required ):
		"""
		Checks if there is enough energy available to perform a certain action.
		Returns 'True' if there is; otherwise 'False' is returned.
		"""

		if self._energy_meter._current_energy >= energy_required:
			return True
		return False
